use druid::{
    widget::{Flex, Label, TextBox},
    AppLauncher,
    Data,
    Lens,
    LensExt,
    PlatformError,
    Widget,
    WidgetExt,
    WindowDesc,
};
use std::{cell::RefCell, rc::Rc};
use util::{ImmutLens, Key};
use uwuifier::{round_up16, uwuify_sse};

mod util;

fn main() -> Result<(), PlatformError> {
    AppLauncher::with_window(WindowDesc::new(build_ui()).title("guwu")).launch(AppData::default())
}

#[derive(Default, Data, Lens, Clone)]
struct AppData {
    input: String,
    output: String,

    uwubuf1: Rc<RefCell<Vec<u8>>>,
    uwubuf2: Rc<RefCell<Vec<u8>>>,
}

impl AppData {
    fn alloc_uwubufs(&self, len: usize) {
        let mut uwubuf1 = self.uwubuf1.borrow_mut();
        let mut uwubuf2 = self.uwubuf2.borrow_mut();

        if len > uwubuf1.len() {
            uwubuf1.resize(len, 0);
        }

        if len > uwubuf2.len() {
            uwubuf2.resize(len, 0);
        }
    }

    fn uwu(&mut self) {
        self.alloc_uwubufs(round_up16(self.input.len()) * 16);

        let mut uwubuf1 = self.uwubuf1.borrow_mut();
        let mut uwubuf2 = self.uwubuf2.borrow_mut();

        let uwuified = uwuify_sse(self.input.as_bytes(), &mut uwubuf1, &mut uwubuf2);

        // SAFETY: uwuify should always return valid UTF-8 if the input string was valid
        // UTF-8, which we can be sure of, as it's a `String`.
        self.output = unsafe { String::from_utf8_unchecked(uwuified.to_vec()) };
    }
}

fn build_ui() -> impl Widget<AppData> {
    Flex::row()
        .with_flex_child(
            Flex::column()
                .with_child(Label::new("Input"))
                .with_flex_child(
                    TextBox::multiline()
                        .with_placeholder("Type here...")
                        .expand()
                        .lens(AppData::input)
                        .controller(Key::new(|_, data: &mut AppData, _, _| data.uwu())),
                    1.,
                ),
            1.,
        )
        .with_spacer(4.)
        .with_flex_child(
            Flex::column()
                .with_child(Label::new("Output"))
                .with_flex_child(
                    TextBox::multiline()
                        .with_placeholder("UwU")
                        .lens(AppData::output.then(ImmutLens))
                        .expand(),
                    1.,
                ),
            1.,
        )
}
